use arti_client::TorClient;
use eyre::Result;
use tokio::io::{AsyncWriteExt, AsyncReadExt};

trait Serialize {
    fn serialize(&self) -> Vec<u8>;
}

#[derive(Clone, Debug)]
pub struct HTTPRequest {
    pub method: HTTPMethod,
    pub location: String,
    pub version: String,
    pub headers: Vec<HTTPHeader>,
    pub body: Option<Vec<u8>>,
}

#[derive(Clone, Debug)]
pub struct HTTPResponse {
    pub version: String,
    pub status_code: u16,
    pub status: String,
    pub headers: Vec<HTTPHeader>,
    pub body: Option<Vec<u8>>,
}

#[derive(Clone, Copy, Debug)]
pub enum HTTPMethod {
    Get,
}

#[derive(Clone, Debug)]
pub struct HTTPHeader {
    pub key: String,
    pub value: String,
}

impl Serialize for HTTPRequest {
    fn serialize(&self) -> Vec<u8> {
        let mut result = vec![];

        let mut method_vec_u8 = self.method.serialize();
        let mut location_vec_u8 = self.location.as_bytes().to_vec();
        let mut version_vec_u8 = self.version.as_bytes().to_vec();
        let mut headers = self
            .headers
            .iter()
            .map(|x| x.serialize())
            .collect::<Vec<_>>()
            .concat();

        result.append(&mut method_vec_u8);
        result.push(b' ');
        result.append(&mut location_vec_u8);
        result.push(b' ');
        result.append(&mut version_vec_u8);
        result.append(&mut headers);
        result.push(b'\n');
        result.push(b'\n');

        if let Some(body) = &self.body {
            let mut body = body.clone();
            result.append(&mut body)
        }

        println!("{}", match String::from_utf8(result.clone()) {
            Ok(v) => v,
            Err(v) => format!("{}", v),
        });
        result
    }
}

impl Serialize for HTTPMethod {
    fn serialize(&self) -> Vec<u8> {
        use HTTPMethod as M;
        match self {
            M::Get => b"GET"
        }.to_vec()
    }
}

impl Serialize for HTTPHeader {
    fn serialize(&self) -> Vec<u8> {
        format!("\n{}: {}", self.key, self.value).into_bytes()
    }
}

pub async fn get_website(
    address: &str,
    port: u16,
    request: HTTPRequest,
    tor_client: TorClient<tor_rtcompat::PreferredRuntime>,
) -> Result<Vec<u8>> {
    let mut stream = tor_client.connect((address, port)).await?;
    stream.write_all(&request.serialize()).await?;
    stream.flush().await?;

    println!("Sent request for {} to {address}", request.location);
    let mut response_status: Vec<u8> = Vec::new();

    loop {
        let response = stream.read_u8().await?;
        response_status.push(dbg!(response));
        if response == b'\n' {
            break;
        }
    }

    let _ = dbg!(String::from_utf8(response_status));

    todo!();
}
